#include <dummy.h>

/*
 * LOD Lights Orchestra Device
 * 
 * v1.0 29/01/2017 (CLOD)
 * v2.0 19/09/2017
 *  - rallongé la longueur de séquence à 30s
 *  - viré le bouton 3 positions
 *  - changement de mode avec le bouton poussoir
 *  - gérer le bouton on/off
 *  - stocke le mode précédent en eeprom
 *  - allongé la taille de la séquence (30s)
 *  - peut enregistrer un pattern même en attendant la connexion au mesh
 *  - nettoyé le code our le rendre plus facilement modifiable
 *  
 * Antoine Schmitt and CLOD Lab, Zinc/SecondeNature
 * 
 */

#include "Sequence.h"
#include <EEPROM.h>

//// NEW
#define BTN_ONOFF D4 // bouton on/off D4
#define BTN_REC D2 // bouton rec
#define THELED D1 // la led
#define LIGHTSENSOR A0 // capteur lumière

// OLD
//#define BTN_REC D4 // bouton rec
//#define THELED D8 // la led
//#define LIGHTSENSOR A0 // capteur lumière

// le mode
#define MODE_OFF 0
#define MODE_DIRECT 1
#define MODE_CONN_LOCAL 2
#define MODE_CONN_DISTANCE 3
uint8_t theMode, modeBeforeOff;

// rec/play mode
#define RECORDING 0
#define PLAYING 1
uint8_t recplayMode = 255; // rien au départ

boolean recInactif; // truc spécial au cas où on enregisttre et qu'on dépasse la fin du buffer..
boolean isConnected; // vrai si on est connecté
boolean isPlayingUser; // vrai si on est en train de jouer un séquence utilisateur

// sequence
#define SEQ_FRAMERATE 40 // frame rate de la séquence
#define SEQ_DURATION_S 30 // durée de la séquence
#define SEQ_MAXINDEX SEQ_DURATION_S*SEQ_FRAMERATE
#define SIG_DURATION_S 1 // dirée du signal
#define SIG_MAXINDEX SIG_DURATION_S*SEQ_FRAMERATE
uint32_t seq_nexttime;
uint32_t seq_deltaT_ms = 1000/SEQ_FRAMERATE;

Sequence * sequence; // la séquence à jouer en boucle
Sequence * sig; // le signal de changement de mode

// gestion du click/double-clic
uint32_t press_time;
uint32_t pressdown_maxdur_ms = 300;
uint32_t pressup_maxdur_ms = 500;
uint32_t press_clicks = 0;
boolean pressIsDown = false;

// timer pour ne pas trop lire le capteur
uint32_t nextPeekTime;

#define MYDEBUG 1
uint32_t debugTime, debugDelay;
boolean debugOn = false;


// --------------------------------------------
// ------- Callbacks principaux
// Appelés par le système lors d'événements importants.
// Pour changer le comportement de la lampe, modifier ce que font ces fonctions.
// --------------------------------------------

// Appelé lorsque le lampe se connecte au réseau
void jeSuisConnecte() {
  // reçu par le réseau
  Serial.printf("CLOD : connecté !! ;-) \n");

  // si je suis en train d'enregsitrer, je ne fais rien
  if (recplayMode == RECORDING) return; // drop

  // si on joue déjà une séquence utilisateur, on laisse, sinon, on met un pattern
  if (isPlayingUser) return;
  
  // reinitalise la séquence avec des pulses lents (2 secondes)
  sequenceTriangle(2000);
  
}

// Appelé lorsque le lampe est déconnectée du réseau
void jeSuisDeconnecte() {
  Serial.printf("CLOD : déconnecté ;-( \n");
  
  // si je suis en train d'enregsitrer, je ne fais rien
  if (recplayMode == RECORDING) return; // drop

  // si on joue déjà une séquence utilisateur, on laisse, sinon, on met un pattern
  if (isPlayingUser) return;
  
  // reinitalise la séquence avec des clignotements lents (1 seconde)
  sequenceSquare(1000);
}

// Appelé lorqu'une séquence est reçue du réseau
void sequenceRecue(uint8_t *seq2, int nbSteps2) {
  Serial.printf("CLOD : séquence reçue ! durée = %ds\n", nbSteps2*seq_deltaT_ms/1000);

  // si je suis en train d'enregsitrer, je ne fais rien
  if (recplayMode == RECORDING) return; // drop
  
  // ?? que faire avec la séquence reçue ?

  // exemple : oublie l'ancienne séquence, la remplace par la nouvelle
  sequence->emptySequence();
  sequence->mixSequence2(seq2, nbSteps2);  
  
  recplayMode = PLAYING; // etre sur qu'on va jouer la séquence
  isPlayingUser = true; // se souvenir que l'on joue une séquence créée par l'humain
}

// Appelé lorsque l'humain commence à enregsitrer une séquence
void debutEnregistrement() {
  Serial.printf("CLOD : début enregistrement : RECORDING\n");
  recplayMode = RECORDING;

  // se mettre au début de la séquence pour effacer l'ancienne
  sequence->rewindSequence();
}

// Appelé lorsque l'humain a fini d'enregsitrer une séquence
void finEnregistrement() {
  Serial.printf("CLOD : fin d'enregistrement : PLAYING\n");
  
  recplayMode = PLAYING;

  // rembobine pour jouer à partir du début
  sequence->rewindSequence();
  isPlayingUser = true; // se souvenir que l'on joue une séquence créée par l'humain
 
  // on envoie aux autres
  envoyerSequence();
}


// ------------------------------------
// -------- commandes disponibles
// voir aussi le fichier Sequence.h pour appeler des commandes directement sur la sequence:
// comme : sequence->rewindSequence();
// ------------------------------------

void sequenceTriangle(uint32_t dur_ms) {
  sequence->patternTriangle(dur_ms*SEQ_FRAMERATE/1000);
  recplayMode = PLAYING;
}
void sequenceSquare(uint32_t dur_ms) {
  sequence->patternSquare(dur_ms*SEQ_FRAMERATE/1000);
  recplayMode = PLAYING;
}
void sequenceSaw(uint32_t dur_ms) {
  sequence->patternSaw(dur_ms*SEQ_FRAMERATE/1000);
  recplayMode = PLAYING;
}
void sequenceFadeOut(uint32_t dur_ms) {
  sequence->patternFadeOut(dur_ms*SEQ_FRAMERATE/1000);
  recplayMode = PLAYING;
}
void sequenceDirac(uint32_t dur_msOn, uint32_t dur_msOff, uint32_t nb) {
  sequence->patternDirac(dur_msOn*SEQ_FRAMERATE/1000, dur_msOff*SEQ_FRAMERATE/1000, nb);
  recplayMode = PLAYING;
}

// envoie la séquence courante à toutes les autres lampes
void envoyerSequence() {
  // seulement si quelque chose à envoyer
  if (sequence->_siz == 0) return;
 
  Serial.printf("CLOD : envoi séquence\n");
  meshSend(bytesToString(sequence->sequence, sequence->_siz));
}


// --------------------------------------------
// ----- Initialisation
// --------------------------------------------

void setup() {
  Serial.begin(115200);

  pinMode(BTN_REC, INPUT_PULLUP);
  pinMode(BTN_ONOFF, INPUT);
  
  pinMode(THELED, OUTPUT);
  recInactif = false;

  adaptativeMinMaxInit(0, 1023);
  nextPeekTime = millis() + 1;
  
  if (MYDEBUG) {
    debugDelay = 20;
    debugTime = millis() + debugDelay;
    debugOn = false;
  }

  sequence = new Sequence(SEQ_MAXINDEX, true);
  sig = new Sequence(SIG_MAXINDEX, false);

  // mode au redémarrage du chip
  // ancien mode : stocké dans eeprom 0
  EEPROM.begin(512);
  uint8_t aMode = EEPROM.read(0);
  EEPROM.end();
  // pour la première fois, n'accepter que 1/2/3, sinon mettre 2
  if (aMode != MODE_DIRECT && aMode != MODE_CONN_LOCAL && aMode != MODE_CONN_DISTANCE) {
    aMode = MODE_CONN_LOCAL;
  }
    
  if (!onoffButtonOn()) {
    // off
    modeBeforeOff = aMode;
    aMode = MODE_OFF;
  }
  setMode(aMode);
  Serial.printf("CLOD START en mode %s\n", (theMode == MODE_DIRECT ? "DIRECT" : (theMode == MODE_CONN_LOCAL ? "MESH LOCAL" : "MESH DISTANT")));

}


// --------------------------------------------
// -------- La Boucle
// --------------------------------------------

void loop() {  
  if (MYDEBUG) {
    debugOn = false; // par défaut
    if (millis() >= debugTime) {
      debugOn = true; // valide sur la totalité de loop
      debugTime = millis() + debugDelay;
    }
  }

  // gestion du on/off
  uint8_t aMode;
  if (!onoffButtonOn()) {
    // off
    if (theMode != MODE_OFF) {
      modeBeforeOff = theMode;
      setMode(MODE_OFF);
    }
  } else {
    // on
    if (theMode == MODE_OFF) {
      setMode(modeBeforeOff);
    }
  }
  

  // gestion du mode OFF
  if (theMode == MODE_OFF) {
    // LED off, et ne rien gérer
    analogWrite(THELED, 0);
    return;
  }
  
  // gestion du click/double-click, etc.. pour changement de mode
  boolean buttonReallyDown = false;
  if (!pressIsDown) {
    // n'était pas down
    if (recButtonOn()) {
      // down now
      press_time = millis();
      pressIsDown = true;
      //Serial.printf("CLOD : btn DOWN. clicks ==%d\n", press_clicks);
      
    } else if (press_clicks > 0) {
      // toujours up, et dejà clické une fois au moins
      if (millis() > press_time + pressup_maxdur_ms) {
        //Serial.printf("CLOD : btn UP TIME OUT. OK. clicks ==%d\n", press_clicks);
        // il s'est passé assez de temps après le dernier clic, on considère qu'il n'y en aura pas de nouveau, on dit que c'est fini
        setMode(press_clicks);
        press_clicks = 0;
      }
    }
  } else {
    // press was down
    if (!recButtonOn()) {
      // up now
      pressIsDown = false;
      if (millis() < press_time + pressdown_maxdur_ms) {
        // si remonté avant le délai max, on considère que c'est un click
        press_clicks++;
        //Serial.printf("CLOD : btn UP. clicks ==%d\n", press_clicks);
        press_time = millis();
      }
              
    } else if (millis() > press_time + pressdown_maxdur_ms) {
      // bouton pressé depuis longtemps : on dit que ce n'est pas un clic et qu'il est vraiment down
      // Serial.printf("CLOD : btn DOWN TIME OUT. REC\n");
      buttonReallyDown = true;
      press_clicks = 0;
    }
  }

 // jouer le signal
  if (!sig->finished()) {
    if (millis() >= seq_nexttime) {
      // récupérer la note dans la sequence, et la jouer sur la LED
      uint8_t lightVal = sig->sequenceGetNextNote();
      
      // les notes sont stockées sur un byte dans la séquence, mais la LED a une sensibilité de 4 bytes.
      uint32_t lightValReal = lightVal*4;
      analogWrite(THELED, lightValReal);
      
      // date de la prochaine note
      seq_nexttime = millis() + seq_deltaT_ms;
    }
  }


  // jouer les modes
  if (theMode == MODE_DIRECT) {
    // mode LIVE
    if (sig->finished()) {
      capteurLumiere();
    }
    return;
  }
  
  if (theMode == MODE_CONN_LOCAL) {
    // mode MODE_CONN_LOCAL : en réseau Mesh
    
    // le réseau fait sa cuisine
    meshLoop();

    // pour rester adaptatif en permanence
    capteurLumiereReadOnly();
 
    // changements de mode recplayMode ?
    
    // vérifier qu'on est toujours connecté.
    if (isConnected && meshConnectionCount() == 0) {
      isConnected = false;
      // il n'y a plus de connection = on a été déconnecté !
      jeSuisDeconnecte();
    }

    // début et fin de l'enregistrement
    if (recplayMode == PLAYING) {
      // je suis en train de jouer
      
      if (buttonReallyDown && !recInactif) {
        // on est en mode play, et le bouton REC est appuyé depuis assez longtemps : on démarre l'enregistrement
        debutEnregistrement();
        
      } else if (recInactif && !recButtonOn()) {
        // on est en train de jouer, on avait forcé la fin d'enregistrement, et le bouton REC vient d'être relaché : on oublie la fin forcée
        // pour que le prochain rec fonctionne
        recInactif = false;
      }
    }

    if (recplayMode == RECORDING && !recButtonOn()) {
      // on est en mode rec et le bouton REC n'est plus appuyé : on stoppe l'enregistrement
      finEnregistrement();
    }

    // ------- LEDs, etc..
    if (recplayMode == PLAYING) {
      // joue la sequence
      // seulement si le signal est fini
      if (sig->finished()) {
        // est-ce l'heure de jouer la prochaine note ?
        if (millis() >= seq_nexttime) {
          // récupérer la note dans la sequence, et la jouer sur la LED
          uint8_t lightVal = sequence->sequenceGetNextNote();
          
          // les notes sont stockées sur un byte dans la séquence, mais la LED a une sensibilité de 4 bytes.
          uint32_t lightValReal = lightVal*4;
          analogWrite(THELED, lightValReal);
          
          // date de la prochaine note
          seq_nexttime = millis() + seq_deltaT_ms;
        }
      }
      
    } else if (recplayMode == RECORDING) {
      // mode RECORDING
      
      // est-ce l'heure d'enregistrer la prochaine note ?
      if (millis() >= seq_nexttime) {

        // récupérer la valeur lumière du capteur
        uint8_t lightVal = capteurLumiere();
        
        // enregistrer valeur dans la séquence
        boolean more = sequence->sequenceSetNextNote(lightVal);
  
        // cas spécial de durée max
        if (!more) {
          // forcer la fin de l'enregistrement : rendre le bouton Rec inactif pour que l'enregistrement ne redémarre pas immédiatement
          // Il sera inactif jusqu'à ce qu'on le relache
          finEnregistrement();
          recInactif = true;
        } else {          
          // prochaine date d'enregistrement
          seq_nexttime = millis() + seq_deltaT_ms;
        }
      }
    }

    return;
  }
}


// ------------------------------------
// ---------- bas niveau
// ------------------------------------

void setMode(uint32_t nbClicks) {
  Serial.printf("CLOD : setMode nb=%d\n", nbClicks);
  if (nbClicks == theMode) {
    Serial.printf("CLOD : setMode already this mode. Dropping.\n");
    return;
  }

  // stop previous mode first
  if (theMode == MODE_CONN_LOCAL) {
    meshStop();
  }

  // start new mode
  if (nbClicks == MODE_OFF) {
    theMode = MODE_OFF;
    
  } else if (nbClicks == MODE_DIRECT) {
    theMode = MODE_DIRECT;
    
  } else if (nbClicks == MODE_CONN_LOCAL) {
    theMode = MODE_CONN_LOCAL;
    isPlayingUser = false;
    isConnected = false;
    recplayMode = PLAYING;
    jeSuisDeconnecte();
    meshStart();
  
/*  } else if (nbClicks == MODE_CONN_DISTANCE) {
    theMode = MODE_CONN_DISTANCE;
    // ??*/
  } else {
    // mode inconnu
    Serial.printf("CLOD : mode inconnu : %d\n", nbClicks);
    return;
  }

  if (theMode != MODE_OFF) {
    // signal pour dire bien reçu
    sig->patternDirac(200*SEQ_FRAMERATE/1000, 200*SEQ_FRAMERATE/1000, nbClicks);
    
    // SAVE to eeprom
    EEPROM.begin(512);
    EEPROM.write(0, theMode);
    EEPROM.end();
  }
}


// ------------------------------------
// -------- gestion des capteurs
// ------------------------------------

// Lecture du bouton REC
boolean recButtonOn()
{
  boolean recOn = (digitalRead(BTN_REC) == 0);
  return recOn;
}

// Lecture du bouton ON/OFF
boolean onoffButtonOn()
{
  boolean isOn = (digitalRead(BTN_ONOFF) == 1);
  return isOn;
}

void capteurLumiereReadOnly() {
  // pour rester adaptatif, que la rétine continue à recevoir de la lumière tout le temps

  // n'aller voir que toutes les 100 ms
  if (millis() < nextPeekTime) return;
  nextPeekTime = millis() + 100;
  
  uint32_t sensorVal = analogRead(LIGHTSENSOR);
  uint32_t lightVal32 = adaptativeMap(sensorVal);
  
  //if (MYDEBUG && debugOn) Serial.printf("CLOD DEBUG : RAW LIGHT = %4d", sensorVal);
  //if (MYDEBUG && debugOn) Serial.printf(" (min=%3d, max=%3d) LED = %4d\n", amm_getCurMin(), amm_getCurMax(), lightVal32);
}

uint8_t capteurLumiere() {
  // avec seuils
  uint8_t lightVal = 0;
  uint32_t sensorVal = analogRead(LIGHTSENSOR);

  //if (MYDEBUG) Serial.printf("CLOD DEBUG : RAW LIGHT = %4d", sensorVal);

  // automap
  uint32_t lightVal32 = adaptativeMap(sensorVal);
  
  // car la note est sur 1 byte
  lightVal = lightVal32/4;
  
  // la LED est sur 4 bytes.
  analogWrite(THELED, lightVal32);
    
  //if (MYDEBUG) Serial.printf(" (min=%3d, max=%3d) LED = %4d", amm_getCurMin(), amm_getCurMax(), lightVal32);
  //if (MYDEBUG) Serial.printf(" note = %3d\n", lightVal);
  
  return lightVal;
}

// -------------------------------------------------------
// ----- communication bas niveau avec le mesh réseau
// -------------------------------------------------------

void receivedCallback( uint32_t from, String &msg ) {
  
  //Serial.printf("mesh: Received from %d msg=%s\n", from, msg.c_str());
  uint8_t sequence2[SEQ_MAXINDEX];
  int receivedBytes;
  stringToBytes(msg, sequence2, SEQ_MAXINDEX, &receivedBytes);

  // debug
  if (MYDEBUG) {
    String printmsg = "Recu "+String(receivedBytes)+" valeurs:";
    for (int i = 0; i < receivedBytes; i++) {
      printmsg += String(sequence2[i]) + ",";
    }
    Serial.printf("mesh: Received from %d printmsg=%s\n", from, printmsg.c_str());
  }

  // traiter
  sequenceRecue(sequence2, receivedBytes);
}

void newConnectionCallback( bool adopt ) {
  int nbConn = meshConnectionCount() ;
  if (MYDEBUG) Serial.printf("mesh: New Connection, adopt=%d + nbconnections=%d\n", adopt, nbConn);
  isConnected = true;
  jeSuisConnecte();
}


